/**************************************************************************

Copyright [2012] [Software Engineering Group - Kiel University]

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

***************************************************************************/

var labelType, useGradients, nativeTextSupport, animate;

/**
	Set up Error Text and Navigator
*/

(function() {
  var ua = navigator.userAgent,
      iStuff = ua.match(/iPhone/i) || ua.match(/iPad/i),
      typeOfCanvas = typeof HTMLCanvasElement,
      nativeCanvasSupport = (typeOfCanvas == 'object' || typeOfCanvas == 'function'),
      textSupport = nativeCanvasSupport 
        && (typeof document.createElement('canvas').getContext('2d').fillText == 'function');
  //I'm setting this based on the fact that ExCanvas provides text support for IE
  //and that as of today iPhone/iPad current text support is lame
  labelType = (!nativeCanvasSupport || (textSupport && !iStuff))? 'Native' : 'HTML';
  nativeTextSupport = labelType == 'Native';
  useGradients = nativeCanvasSupport;
  animate = !(iStuff || !nativeCanvasSupport);
})();

var Log = {
  elem: false,
  write: function(text){
    if (!this.elem) 
      this.elem = document.getElementById('log');
    this.elem.innerHTML = text;
    this.elem.style.left = (500 - this.elem.offsetWidth / 2) + 'px';
  }
};


var testVar = "no";

function GraphFlow(){
	/////////////////////////////////////////////
	//				VARIABLES				   //
	/////////////////////////////////////////////
	
	var visContainer = 'infovis';
	
	/** Colors */
	var nodeFillColor = [];
		nodeFillColor['nodeFamily']		= "#DEDEDE";
	
	var nodeStrokeColor = [];
		nodeStrokeColor['inputPort']   	  = "#AA0000";
		nodeStrokeColor['outputPort']	  = "#AA0000";
		nodeStrokeColor['repositoryPort'] = "#AA0000";
		nodeStrokeColor['nodeFamily'] 	  = "#4D4D4D";
		nodeStrokeColor['crossBox'] 	  = "#4D4D4D";
		
	var nodeColorFocus	  = "#0098BE",
		edgeColor 		  = "#114270",
		edgeColorFocus 	  = "#0098BE";
		
	/** Icon Images */
	var nodeIcons = [];
	
	/** Overall size modifier: the width of node labels in pixels */
	var vertexLabelSize 	= 12;

	/** FlowGraph Object */
	var fd;

	/** Nodes and edges that are highlighted when the mouse moves over them */
	var hover = null;
	var markedNode = null;

	/** This array stores all Nodes and is initialized with a dummy,
		used for drawing edges between a node and the mousepointer. Empty
		elements are dynamically reserved for additional nodes.
		*/
	var json;
		
	/** Used for dynamically adding and removing space for node creation/deletion
		resulting in better performance when manipulating the json array.
	*/
	var jsonCapacity = {'occupied' : 0,
						'max'  : 10};

	/** selected node:	{id : the id of the selected node, 
						 from : the data.$type of the selected node,
						 label: the label of a selected edge}
						 
		When clicking a node or edge, this variable will contain node
		information, that is required for attaching/detaching edges.
		*/
	var selectedNode = null;
	
	/** This points to a DOM-node (not the json-array node!) which is
		an invisible dummy for moving an edge with the mouse
		*/
	var mouseNode = null;

	/** true if the mouse is over a port/edge */
	var mouseOverNode = false;
	var mouseOverEdge = false;
	
	/** A hashtable, mapping mouse and node events to arrays of functions that
		are called when these events happen
		*/
	var listener = [];
	
	/** Can be changed to allow or prohibit node movement or deletion/creation */
	var readOnly = {'deleteCreate' 	: true,
					'move' 			: true};
	
	/** Grid dummy-node which is drawn over the graph if it is visible */
		
	var grid = { 'data': {'$dim' : vertexLabelSize*4,
						  '$type'	 : 'grid',
						  '$color'  : '#AAAAFF',
						  '$width' : document.getElementById(visContainer).clientWidth,
						  '$height': document.getElementById(visContainer).clientHeight
						},
				 'id':	'#DUMMY_GRID_NODE',
				 'visible': false,
				 'snap'   : false
				};
	
	/** Stores information about the canvas, such as its dimensions and its position.
		This information is needed to correctly draw the Grid and may come in handy in
		future functions.
		*/
	var navi = { // screen position
				 'centerX' : 0,
				 'centerY' : 0,
				 // these determine when the screen should move with a dragged node
				 'borderLeft' : 0,
				 'borderRight' : 0,
				 'borderTop' : 0,
				 'borderBottom' : 0,
				 // memorize where we touch a node upon dragging
				 'nodeOffsetX' : 0,
				 'nodeOffsetY' : 0,
				 // how far a node will be dragged
				 'dragX' : null,
				 'dragY' : null,
				 // dragged nodefamily and ports
				 'draggedNodes' : null};
	
	/////////////////////////////////////////////
	//				FUNCTIONS				   //
	/////////////////////////////////////////////

	/**
	 * Measures the width of a text on the canvas
	 * 
	 * @param text the text that is measured
	 * @param font the size and font of the text
	 * @returns the measured width
	 */
	this.getTextWidth = function(text, font){
		
		// validate arguments
		if(!validArg("getTextWidth()", text, "string") ||
		   !validArg("getTextWidth()", font, "string")){
			return;
		}
			
		// get canvas context
		var ctx = fd.canvas.getCtx(),
			width = 0;
		
		// measure width
		ctx.save();
		ctx.font = font;
		width = ctx.measureText(text).width;
		ctx.restore();
		
		return width;
	}
	
	/**
	 * @returns the size modifier that scales the whole graph.
	 */
	this.getSizeModifier = function(){ 
		return vertexLabelSize;
	}
	
	/**
	 * Adds a node (that is NOT a Nodefamily) to the graph.
	 * 
	 * @param xPosition - the horizontal position of the node
	 * @param yPosition - the vertical position of the node
	 * @param node - an object which at least contains an 'id' and may have a 'data' field 
	 * @param type - the type of the defined custom node (basic types are: inputPort, nodeFamily, ...)
	 * @param noPlot - if false, a refresh is required to apply the changes
	 * @param forced - if true, does not throw a listener event
	 */
	this.addCustomNode = function(xPosition, yPosition, node, type, noPlot, forced){
		
		// validate arguments
		if(!validArg("addCustomNode()", xPosition, "number") ||
		   !validArg("addCustomNode()", yPosition, "number") ||
		   !validArg("addCustomNode()", type, "string") ||
		   !validArgNode("addCustomNode()", node)){
			return;
		}
		
		var nodeType = 'Custom';
		
		if(node.data && node.data.$nodeType){
			nodeType = node.data.$nodeType;
		}
		
		// define node
		var newNode = {
				  "adjacencies": [], 
				  "data": {
					"$dim": vertexLabelSize,					
					"$type": type,
					"$icon": nodeIcons[nodeType],
					"$nodeType": nodeType,
					"$xPos": xPosition,
					"$yPos": yPosition,
					"$custom": true
				  }, 
				  "id": node.id, 
				  "name": node.name};
		
		// add colors if they exist
		var	strokeColor = nodeStrokeColor[type],
			nodeColor = nodeFillColor[type];
		if(strokeColor){ newNode.data.$color = strokeColor;}
		if(nodeColor){ newNode.data.$fillColor = nodeColor;}
		
		// (over-)write data if it was specified in node.data
		if(node.data){
			for(var field in node.data){
				newNode.data[field] = node.data[field];
			}
		}
		
		// add node to json array
		json.push(newNode);
		
		// apply changes immediately
		if(!noPlot){
			fd.graph.addNode(newNode);
			fd.graph.getNode(newNode.id).pos.setc(xPosition, yPosition);
			fd.plot();
		}
		
		// call listener
		if(!forced){
			callListener("onCreateNode", [newNode]);
		}
	}
	
	/**
	 * Zooms in or out, depending on the factor
	 * 
	 * @param factor -  the graph zooms out if it is smaller than 1
	 */
	this.zoom = function(factor){
		
		// check argument
		if(!validArg("zoom()", factor, "number")){
			return;
		}else if(factor < 0){
			callListener("onError", ["Error in function zoom(): Argument factor must be a positive number!"]);
		}
		
		// zoom
		var canvas = fd.canvas;
		canvas.scale(factor, factor);
		
		// show text only if it is big enough
		canvas.showLabels = (canvas.scaleOffsetX > canvas.labelThreshold);
	}
	
	/**
		Checks if the GraphFlow container has changed in size and
		updates the canvas area accordingly.
	*/
	this.updateCanvasSize = function(){
		var container = document.getElementById(visContainer);
		var widthOld = grid.data.$width,
			heightOld = grid.data.$height,
			widthNew = container.clientWidth,
			heightNew = container.clientHeight;
			
		// check if the dimensions changed
		if( widthOld != widthNew || heightOld != heightNew){
		
			// calculate translation parameters
			var scale = fd.canvas.scaleOffsetX,
				transX = fd.canvas.translateOffsetX + (widthOld - widthNew)/2,
				transY = fd.canvas.translateOffsetY + (heightOld - heightNew)/2;
			
			// resize canvas, scale and translate content to maintain its position
			fd.canvas.resize(container.clientWidth, container.clientHeight, true);
			fd.canvas.scale(scale, scale);
			fd.canvas.translate(transX / scale, transY / scale);
			
			// update grid dimensions
			grid.data.$width = container.clientWidth;
			grid.data.$height = container.clientHeight;
			
			// apply changes to the grid if visible
			if(grid.visible){
				var domGrid = fd.graph.getNode(grid.id);
				domGrid.data.$width = widthNew;
				domGrid.data.$height = heightNew;
			}
		}
		fd.plot();
	}
	
	/**
		Concatenates nodefamily 'id's and their x- and y-position
		to a string, separating each element with a space.
		TODO: save bend points of edges
		@return - the resulting string
	*/
	this.saveNodePositions = function(){
		var positions = "";
		var node;
		
		for(var n = 0, l = jsonCapacity.occupied; n < l; n++){
			node = json[n];
			positions += node.id + " " + node.data.$xPos + " " + node.data.$yPos + " ";
		}
		
		return positions;
	}
	
	/**
		Changes one or more properties of an edge immediately.
		@param sourceID - the ID of the node where the edge starts
		@param targetID - the ID of the node where the edge ends
		@param properties - a data object that may contain any field from
					  data as well as "id". Data fields must start with a $.
		@param noPlot - if true, the graph will not be plotted anew
	*/
	this.setEdgeData = function(sourceID, targetID, properties, noPlot){
		
		// check arguments
		if(!validArg("setEdgeData()", sourceID, "string") ||
		   !validArg("setEdgeData()", targetID, "string")){
			return;
		}
		
		var node = getNode(sourceID),
			domEdge = fd.graph.getAdjacence(sourceID, targetID),
			edge;
		
		// find edge in json-array
		for(var e in node.adjacencies){
			if(e.nodeTo == targetID){
				edge = e;
				break;
			}
		}
		
		// iterate through properties object
		var value;
		for(var field in properties){
			value = properties[field];
			
			// make changes to data
			if(field[0] == '$'){
				
				if(field == '$type'){
					// apply new value to data
					edge.Edge[field] = value;
					if(!noPlot){ domEdge.Edge[field] = value;}
				}
				else{
					// apply new value to data
					edge.data[field] = value;
					if(!noPlot){ domEdge.data[field] = value;}
				}
			}
			
			// make changes to name or id
			else{
				edge[field] = value;
				if(!noPlot){ domEdge[field] = value;}
			}
		}
		
		// apply visual changes
		if(!noPlot){ fd.plot();}
	}
	
	/**
		Changes one or more properties of a node immediately.
		@nodeID - the ID of the node that needs changing
		@properties - a data object that may contain any field from
					  node.data as well as "name" and "id". Data fields
					  must start with a $.
		@noPlot - if true, the graph will not be plotted anew
	*/
	this.setNodeData = function(nodeID, properties, noPlot){
		
		// check argument
		if(!validArg("setNodeData()", nodeID, "string")){
			return;
		}
		
		var value;
		var node = getNode(nodeID),
			domNode = fd.graph.getNode(nodeID);
		
		// iterate through properties object
		for(var field in properties){
			value = properties[field];
			
			// make changes to data
			if(field[0] == '$'){
				
				// treat a new font name by appending the font size as well
				if(field == "$font"){
					if(!value){
						value = "Lucida Console";
					}
					value = "px " + value;
				}
				// apply new value to data
				node.data[field] = value;
				if(!noPlot){ domNode.data[field] = value;}
			}
			
			// make changes to name or id
			else{
				node[field] = value;
				if(!noPlot){ domNode[field] = value;}
			}
		}
		
		// recalculate node dimensions in case we changed the font or name
		if(!node.data.$custom){
			updateNodeWidth(node, domNode);
		}
		
		// apply visual changes
		if(!noPlot){ fd.plot();}
	}
	
	/**
		Recalculates the width of a node and repositions the ports.
		Applies the change immediately.
		@param jsonNode - the node in the json array
		@param domNode - the node of the actual displayed graph
	*/
	function updateNodeWidth(jsonNode, domNode){
		var data = jsonNode.data;
		
		// reserve icon space
		var iconSpace = 0;
		if(nodeIcons[data.$nodeType]){iconSpace = 6;}
		
		// calculate width of name
		var width = (jsonNode.name.length+4+iconSpace) * vertexLabelSize;
		
		// check if classname is longer than width
		if(data.$nodeClass){
			var classWidth = (data.$nodeClass.length+6+iconSpace)*(vertexLabelSize-2);
			width = Math.max(width, classWidth);
		}
		
		// update width
		jsonNode.data.$width = width;
		domNode.data.$width = width;
		
		width /= 2;
		
		// update port positions
		var port, domPort, relX, pos, offset, x, y;
		for(var p = data.$portIndex+ jsonCapacity.max, l = p + data.$portCount; p < l; p++){
			port = json[p];
			domPort = fd.graph.getNode(port.id);
			if(port.data.$relativeX){
				relX = port.data.$relativeX;
				if(relX < 0){
					offset = -width;
				}else{
					offset = width;
				}
				
				// update relative position
				port.data.$relativeX = offset;
				domPort.data.$relativeX = offset;
			}
			
			// update position
			pos = domPort.pos.getc(true);
			x = pos.x+offset-relX;
			y = pos.y;
			domPort.pos.setc(x, y);
			port.data.$xPos = x;
			port.data.$yPos = y;
		}
	}
	
	/**
		Reads nodefamily positions out of a string, which was created with
		the saveNodePositions() function and moves all nodes to the respective
		positions.
		@param positions - the string which contains node ids and positions
		@param smoothTransition - no functionality at the moment
	*/
	this.loadNodePositions = function(positions, smoothTransition){
		var pos = positions.split(" ");
		var node, x, y;
		
		for(var i = 0, l = pos.length; i < l; i+=3){
			// look for the node
			node = fd.graph.getNode(pos[i]);
			
			// if node exists, set positions
			if(node){
				x = parseFloat(pos[i+1]);
				y = parseFloat(pos[i+2]);
				
				prepareNodeMove(node);
				moveNode(x,y);
				saveNodeMove();
			}
		}
		
		//TODO: smooth animation
		if(false){
			fd.animate({  
				modes: ['linear'],
				transition: $jit.Trans.Cubic.easeInOut,
				duration: 1000
			}); 
		}
		else{
			fd.plot();
		}
	}
	
	/**
		Extracts certain node and edge information and fuses it to a single
		String, which will be argument of a listener event and can be used
		by an autolayouting algorithm.
	*/
	this.autoLayout = function(){
		
		// this array will contain XMI-Strings
		var indexMap = [];		// KGraphs use indices instead of IDs
		var n, l;
		
		// determine node indices
		for(n = 0, l = jsonCapacity.occupied; n < l; n++){
			indexMap[json[n].id] = n;
		}
		
		// determine port indices
		var node, indexCounter = 0;
		for(n = jsonCapacity.max, l = json.length; n < l; n++){
			node = json[n];
			if(node.data.$type == "crossBox"){
				indexCounter = 0;
			}
			else{
				indexMap[node.id] = indexCounter;
				indexCounter++;
			}
		}
		
		// prepare lots of loop variables
		var nodes = ""+ vertexLabelSize,
			edges = "";
		
		var data, p, pl, a, al, height;
		var repoCount, inputCount, outputCount;
		var repoY = 0, inputY = 0, outputY = 0;
		
		// now convert the json array to two simple integer arrays
		for(n = 0, l = jsonCapacity.occupied; n < l; n++){
			data = json[n].data;
			nodes += " " + data.$width;
			height = data.$height;
			nodes += " " + height;
			height /= 2;
			
			// count ports
			inputCount = 0;
			repoCount = 0;
			outputCount = 0;
			for(p = jsonCapacity.max + data.$portIndex+1, pl = p + data.$portCount -1; p < pl; p++){
				data = json[p].data;
				switch(data.$type){
					case "repositoryPort" : 
						repoCount ++; 
						if(data.$relativeY){
							repoY = data.$relativeY + height;
						}
						break;
					case "inputPort" : 
						inputCount ++; 
						if(data.$relativeY){
							inputY = data.$relativeY + height;
						}
						break;
					case "outputPort" : 
						outputCount ++; 
						if(data.$relativeY){
							outputY = data.$relativeY + height;
						}
						break;
				}
				
				// check outgoing edges
				adja = json[p].adjacencies;
				if(adja){
					for( a = 0, al = adja.length; a < al; a++){
						node = adja[a];
						edges += n + " "
								+ indexMap[node.data.$targetFamily] + " "
								+ indexMap[node.nodeFrom] + " "
								+ indexMap[node.nodeTo] + " ";
					}
				}
			}
			nodes += " " +inputCount + " " 
						+ inputY + " " 
						+ repoCount + " " 
						+ repoY + " " 
						+ outputCount + " " 
						+ outputY;
		}
		
		callListener("autoLayout", [nodes, edges]);
	}
	
	/**
		Assumes that the graph has not been altered since the last call of autolayout().
		Reads a string containing new x,y positions, which are separated by spaces and
		moves nodefamilies to the specified positions.
		
		@param positions - the position string
	
	*/
	this.loadPositionsFromLayout = function(input){
		// split string by spaces
		var bendPoints, positions;
		{
			var inputSplit = input.split("#");
			positions = inputSplit[0].split(" ");
			
			// bend points are optional
			if(inputSplit[1] && inputSplit[1].length > 0){
				bendPoints = inputSplit[1].split(" ");
			}

		}
		// prepare loop variables
		var node, data, adja, x, y, p = 0;
		
		var xOffset = -grid.data.$width/2,
			yOffset = -grid.data.$height/2;
		
		////
		// Node Positions
		
		// iterate nodefamiles
		for(var n = 0, l = jsonCapacity.occupied; n < l; n++){
			// extract positions from string array
			x = parseFloat(positions[p++]) + xOffset;
			y = parseFloat(positions[p++]) + yOffset;
			
			// find node in graph and move it
			// we do it in this complicated way to
			// make smooth node positioning possible
			node = fd.graph.getNode(json[n].id);
			prepareNodeMove(node);
			moveNode(x,y);
			saveNodeMove();
			data = json[n].data;			
		}
		
		////
		// BendPoints
		if(bendPoints){
			
			var adja, a, al
				e = 0,
				bendList = [];
			make_bendPoints:{
				// look for ports with edges
				for(var po = jsonCapacity.max+1, l = json.length; po < l; po++){
					adja = json[po].adjacencies;
					if(adja){
						// iterate through outgoing edges
						for(a = 0, al = adja.length; a < al; a++){
							if(e == bendPoints.length){
									break make_bendPoints;
							}
							
							// read from bend point string
							while(bendPoints[e] != ";" && bendPoints[e] != ""){
								x = parseFloat(bendPoints[e++]) + xOffset;
								y = parseFloat(bendPoints[e++]) + yOffset;
								bendList.push({'x':x, 'y':y});
							}
							if(bendList.length > 0){
								setBendPoints(adja[a].nodeFrom, adja[a].nodeTo, bendList);
								// prepare a new set of bend points
								bendList = [];
							}
							// increment if we have a separating semicolon
							e++;
						}
					}
				}
			}
		}
		// may be animated in the future
		fd.plot();
	}
	
	/**
		Checks an input color-string for validation.
		Sends an onError-event if the string is not valid.
		@param callFunction - the function where the argument needs to be checked
		@param color - the color argument which is checked
		@returns - false if the string is not a well formed color string 
	*/
	this.validArgColor = function(callFunction, color){
		var valid = false;
		
		// is it a string at all ?
		if(typeof color == "string"){
			// does the string match "#RRGGBB" ?
			var regEx = /^#[a-f\d]{6}$/i;
			valid = color.match(regEx);
		}
		
		if(!valid){
			callListener("onError", ["Error in function "+ callFunction 
										+ ": Invalid color '"+ color 
										+ "'. Must be structured like '#RRGGBB'."]);
		}
		return valid;
	}
	
	
	/**
		Checks an input object for validation.
		Sends an onError-event if the Object does not have an id
		@param callFunction - the function where the argument needs to be checked
		@param node - the object argument which is checked
		@returns - false if the object does not have an id
	*/
	this.validArgNode = function(callFunction, node){
		var valid = false;
		
		// is it an object at all ?
		if(node && typeof node == "object"){
			// does the object have an id
			valid = (typeof node.id != "undefined");
		}
		
		if(!valid){
			callListener("onError", ["Error in function "+ callFunction 
										+ ": Invalid Node '"+ node 
										+ "'. Must have a field 'id'."]);
			return valid;
		}
		
		var id = node.id; 
		/*
		findDuplicate:{
			for(var n = 0, l = jsonCapacity.occupied; n < l; n++){
				if(json[n].id == id){
					valid = false;
					break findDuplicate;
				}
			}
			
			for(var n = jsonCapacity.max, l = json.length; n < l; n++){
				if(json[n].id == id){
					valid = false;
					break findDuplicate;
				}
			}
		}
		
		
		if(!valid){
			callListener("onError", ["Error in function "+ callFunction 
										+ ": Invalid Node '"+ node 
										+ "'. The ID '"+ id +"' is already taken."]);
		}
		*/
		return valid;
	}
	
	
	/**
		Checks an argument for type correctness.
		Sends an onError-event if the argument is not valid.
		@param callFunction - the function where the argument needs to be checked
		@param arg - the argument which is checked
		@param type - the type of which the argument must be
		@returns - false if the string is not a well formed color string 
	*/
	this.validArg = function(callFunction, arg, type){
		var valid = (typeof arg == type);
		if(!valid){
			callListener("onError", ["Error in function "+ callFunction 
										+ ": Invalid argument type for argument '"
										+ arg + "'. Must be of type " + type + "."]);
		}
		return valid;
	}
	
	/**
		Changes the color of the grid.
		@param newColor - the new color of the grid
		@param noPlot - if true, the graph will not be plotted anew
	*/
	this.setGridColor = function(newColor, noPlot){
		if(validArgColor('setGridColor()', newColor)){
			grid.data.$color = newColor;
			
			// apply changes immediately
			if(!noPlot && grid.visible){
				var domGrid = fd.graph.getNode(grid.id);
				domGrid.data.$color = newColor;
				
				fd.plot();
			}
		}
	}
	
	/**
		Changes the size of the grid.
		@param newSize - the new size of the grid
		@param noPlot - if true, the graph will not be plotted anew
	*/
	this.setGridSize = function(newSize, noPlot){
		if(validArg('setGridSize()', newSize, 'number')){
			grid.data.$dim = newSize;
			
			// apply changes immediately
			if(!noPlot &&grid.visible){
				var domGrid = fd.graph.getNode(grid.id);
				domGrid.data.$dim = newSize;
				
				fd.plot();
			}
		}
	}
	
	/**
		Changes whether Nodes should snap to grid.
		@param snap - must be true if Nodes should snap
	*/
	this.setGridSnap = function(snap){
		grid.snap = snap;
	}
	
	/**
		Changes the visibility of the grid.
		@param visibility - must be true to display the graph
		@param noPlot - if true, the graph will not be plotted anew
	*/
	this.setGridVisible = function(visibility, noPlot){
		grid.visible = visibility;
		
		// apply changes immediately
		if(!noPlot){
			if(visibility){
				fd.graph.addNode(grid);
			}
			else{
				fd.graph.removeNode(grid.id);
			}
			// apply visual changes
			fd.plot();
		}
	}
	
	/**
		Scales and centers the graph to fit the space provided by the container
		in which it is drawn.
	*/
	this.scaleToFit = function(){
	// do nothing if the graph is empty
	if(jsonCapacity.occupied == 0){
		return;
	}
	
	// init loop variables
	var data   = json[0].data,
		width  = data.$width/2,
		height = data.$height/2;
	
	var left  	= data.$xPos - width,
		right 	= data.$xPos + width, 
		top		= data.$yPos - height, 
		bottom	= data.$yPos + height;
	
	// init the outer bounds of the displayed graph
	var leftMost = left,
		rightMost = right,
		topMost = top,
		bottomMost = bottom;
	
	// check nodeFamilies for position
	for(var n = 1, l= jsonCapacity.occupied; n < l; n++){
		data   = json[n].data,
		width  = data.$width/2,
		height = data.$height/2;
	
		left  	= data.$xPos - width,
		right 	= data.$xPos + width, 
		top		= data.$yPos - height, 
		bottom	= data.$yPos + height;
		
		// update bounds
		if(left < leftMost){
			leftMost = left;
		}
		if(right > rightMost){
			rightMost = right;
		}
		if(top < topMost){
			topMost = top;
		}
		if(bottom > bottomMost){
			bottomMost = bottom;
		}
	}
	// check custom nodes for position
	for(var n = jsonCapacity.max+1, l= json.length; n < l; n++){
		data   = json[n].data;
		
		if(data.$custom){
			width  = data.$width/2,
			height = data.$height/2;
		
			left  	= data.$xPos - width,
			right 	= data.$xPos + width, 
			top		= data.$yPos - height, 
			bottom	= data.$yPos + height;
			
			// update bounds
			if(left < leftMost){
				leftMost = left;
			}
			if(right > rightMost){
				rightMost = right;
			}
			if(top < topMost){
				topMost = top;
			}
			if(bottom > bottomMost){
				bottomMost = bottom;
			}
		}
	}
	
	// scaling values
	var canvas = fd.canvas,
		oldScale = canvas.scaleOffsetX,
		scaleX = grid.data.$width / (rightMost - leftMost + 4*vertexLabelSize),
		scaleY = grid.data.$height / (bottomMost - topMost + 4*vertexLabelSize)
		scale = Math.min(scaleX, scaleY) / oldScale;
		
	// translation values
	var midX = (leftMost + rightMost) / 2,
		midY = (topMost + bottomMost) / 2 -vertexLabelSize;
		
	
	// translate to origin, scale, and translate to the center of the graph
	canvas.translate( -canvas.translateOffsetX/oldScale, -canvas.translateOffsetY/oldScale);
	canvas.scale(scale, scale);
	canvas.translate( -midX, -midY);
	
	// show text only if it is big enough
	canvas.showLabels = (canvas.scaleOffsetX > canvas.labelThreshold);
	
	fd.plot();
	}
	
	
	/**
	Assigns an iconPath to a type of node. Needs a refresh() to take effect.
	@param nodeType - the type of the node. Typically 'Repository',
					  'Filter', or 'Reader'
	@param path - the path to the image. If null, an existing
				  icon will be deleted
	*/
	this.setNodeIcon = function(nodeType, path){
		if(!validArg('setNodeIcon()', nodeType, 'string')){
			return;
		}
	
		if(path == null){
			delete nodeIcons[nodeType];
			
			// remove icons from displayed graph and shrink nodes
			var jsonNode, domNode, data;
			for(var n = 0, l = jsonCapacity.occupied; n < l; n++){
				
				// get data
				jsonNode = json[n];
				data = jsonNode.data;
				
				// update node width and port positions
				if(data.$nodeType == nodeType){
					
					domNode = fd.graph.getNode(jsonNode.id);
					delete data.$icon;
					delete domNode.data.$icon;
					updateNodeWidth(jsonNode, domNode);
				}
			}
			fd.plot();
		}
		else{
			if(!validArg('setNodeIcon()', path, 'string')){
				return;
			}
			var icon;
			icon = new Image();
			
			// load error
			icon.onerror = function(){
				callListener("onError", ["Error in function setNodeIcon(): Could not load '"+ path +"'."]);
			};
			
			// load success
			icon.onload = function(){
				// this may happen on IE, when linking to images via web
				if(!icon.complete){
					callListener("onError", ["Error in function setNodeIcon(): Timeout while loading '"+ path +"'. Please try again!"]);
					return;
				}
				
				// if nodes were already created, we need to
				// adjust their width and iconPath
				var prevImage = nodeIcons[nodeType];
				nodeIcons[nodeType] = icon;
				
				if(!prevImage){
					
					var jsonNode, domNode, data;
					for(var n = 0, l = jsonCapacity.occupied; n < l; n++){
						
						// get data
						jsonNode = json[n];
						data = jsonNode.data;
						
						// update node width and port positions
						if(data.$nodeType == nodeType){
							
							domNode = fd.graph.getNode(jsonNode.id);
							data.$icon = icon;
							domNode.data.$icon = icon;
							updateNodeWidth(jsonNode, domNode);
						}
					}
				}
				fd.plot();
			};
			
			// attempt to load the image
			icon.src = path;
		}
	}
	
	/**
	Changes the appearance of the mouse cursor. 
	@param newCursor - 	a String-name of the new cursor. 
						If null, restore the cursor.
	*/
	this.setMouseCursor = function(newCursor){
		if(newCursor == null){
			fd.canvas.getElement().style.cursor = '';
			
		}else if(validArg('setMouseCursor()', newCursor, 'string')){
			fd.canvas.getElement().style.cursor = newCursor;
		}
	}
	
	
	/** 
		Reloads the graph, restoring all positions and displaying all
		changes to the graph that happened since the last refresh.
		This function needs to be called whenever nodes or egdges are
		added or deleted!
	*/
	this.refresh = function(){
		// reload graph
		var mouseIndex = jsonCapacity.max;
		
		fd.loadGraphFlowJSON(json, mouseIndex);
		
		if(grid.visible){
			fd.graph.addNode(grid);
		}
		fd.plot();
		
		// set the mouseNode to the freshly created graph node
		mouseNode = fd.graph.getNode(json[mouseIndex].id);
	}
	
	/**
		Adds a functionality to the following MouseEvents:
			onClick,
			onRightClick,
			onMouseMove,
			onMouseEnter,
			onMouseLeave,
			onDragStart,
			onDragMove,
			onDragCancel,
			onDragEnd
		Or one of these events:
			onError(errorMessage)
			onCreateEdge(sourceNode, targetNode, sourcePort, targetPort) : boolean,
			onCreateNode(node) : boolean,
			onRemoveEdge(sourceNode, targetNode, sourcePort, targetPort) : boolean,
			onRemoveNode(node) : boolean
		@param eventName - the name of the mouse event ( see list above)
		@param listenerFunction - a function that may alter the graph
	*/
	this.addListener = function(eventName, listenerFunction){
		// check arguments
		if(!validArg('addListener()', eventName, 'string') ||
			!validArg('addListener()', listenerFunction, 'function')){
				return;
		}
	
		if(listener[eventName] == undefined){
			listener[eventName] = [];
		}
		listener[eventName].push(listenerFunction);
	}
	
	/**
		Calls all listeners that are registered under the eventName.
		@param eventName - see 'addListener()'
		@param arguments - an array of arguments. For mouseEvents it is [node, eventInfo, e],
						   for nodeEvents it is [node], 
						   and for edgeEvents it is [sourceNode, targetNode, sourcePort, targetPort]
		@return true if an editor action is permitted
	*/
	function callListener(eventName, arguments){
		var lArr = listener[eventName];
		
		if(!lArr){
			return true;
		}
		var permitted = true;
		var testval;
		
		// iterate through all listener this event concerns
		for(var e in lArr){
			testval = lArr[e].apply(lArr[e], arguments);
			
			if(testval == false){
				permitted = false;
			}
		}
		return permitted;
	}
	
	/**
	Adds a listener that prohibits edge creation if...
		... an edge with the same source and target exists.
		...	the target node is the source node.
		... the source and target node are of the same port type.
		... a repository node is not connected to a repository and
			vice versa.
	*/
	this.addEdgeConstraints = function(){
		addListener("onCreateEdge", 
		function(sourceFamily, targetFamily, sourcePort, targetPort){
			// remove Edge if it does not lead to an inputPort
			if(targetPort.data.$type != "inputPort"){
				return false;
			}
		
			// remove Edge if the connected nodes share the same type
			if(sourcePort.data.$type == targetPort.data.$type){
				return false;
			}
			
			var goesToRepo = (targetFamily.data.$nodeType == "Repository"),
				isRepoPort = (sourcePort.data.$type == "repositoryPort");
			
			// remove Edge if it is falsely connected to repository ports
			if(goesToRepo != isRepoPort){
				return false;
			}
			
			// remove duplicate Edge
			var adja = sourcePort.adjacencies,
				duplicate = false;
			for(var a=0; a < adja.length; a++){
				if(adja[a].nodeTo == targetPort.id){
					if(duplicate){
						return false;
					}
					else{
						duplicate = true;
					}
				}
			}
			return true;
		});
	}
	
	/**
	Iterates through all nodes and ports, calling a given function for each one.
	@param nodeFunction - a single parameter function, called on a node.
	*/
	this.iterateAllNodes = function(nodeFunction){
		// check arguments
		if(validArg('iterateAllNodes()', nodeFunction, 'function')){
			return;
		}
	
		var node;
		for(var n=0, l=jsonCapacity.occupied; n < l; n++){
			node = json[n];
			nodeFunction(node);
		}
		for(var n=jsonCapacity.max+1, l= json.length; n < l; n++){
			node = json[n];
			nodeFunction(node);
		}
	}
	
	/**
	Changes the colors of the entire graph.
	Requires a refresh().
	@param nodeType - the name of the nodeType, whose fill and stroke color are changed
	@param fillColor - the fill color of nodes. If null, restore Node Color from global Array.
	@param strokeColor - the stroke color of nodes. If null, restore Stroke Color from global Array.
	@param edgeColor - the color of edges. If null, restore Edge Color from global Array.
	@param highlightColor - the of highlighted objects. If null, restore Edge Color from global Array.
	*/
	this.setNodeStyle = function(nodeType, fillColor, strokeColor, edgeColor, focusColor){
		if(validArg('setNodeStyle()', nodeType, 'string')){
		
			if(fillColor && validArgColor('setNodeStyle()', fillColor)){
				nodeFillColor[nodeType] = fillColor;
			}
			if(strokeColor && validArgColor('setNodeStyle()', strokeColor)){
				nodeStrokeColor[nodeType] = strokeColor;
			}
		}
		
		if(edgeColor && validArgColor('setNodeStyle()', edgeColor)){
			edgeColor = edgeColor;
		}
		if(focusColor && validArgColor('setNodeStyle()', focusColor)){
			edgeColorFocus = focusColor;
			nodeColorFocus = focusColor;
		}
		
		
		var data, a, al, adja;
		// change family colors
		if(nodeType == "nodeFamily"){
			for(var n = 0, l = jsonCapacity.occupied; n < l; n++){
				data = json[n].data;
				// change only if node is not marked
				if(json[n] != markedNode){
					data.$fillColor = nodeFillColor["nodeFamily"];
					data.$color = nodeStrokeColor["nodeFamily"];
				}
			}
		}
		else{
			// change colors of ports or custom nodes
			for(var n = jsonCapacity.max+1, l = json.length; n < l; n++){
				data = json[n].data;
				
				if(data.$type == nodeType){
					data.$fillColor = nodeFillColor[nodeType];
					data.$color = nodeStrokeColor[nodeType];
				}
				// change edge colors
				if(edgeColor){
					adja = json[n].adjacencies; // TODO: data.adjacencies?
					for(a = 0, al = adja.length; a < al; a++){
						adja[a].data.$color = edgeColor;
					}
				}
			}
		}
	}
	
	/**
	Permits or prohibits node and edge deletion/creation and hides the deletion-cross
	accordingly. Requires a refresh() to hide/show the deletion-crosses.
	@param deleteCreate - if true, one can change the graph in any way
	@param move - if true, one can change the graph in any way
	*/
	this.setReadOnly = function(deleteCreate, move){
				
		if(deleteCreate != readOnly.deleteCreate){
			// set crossBoxes (in)visible
			var data;
			for(var n=jsonCapacity.max+1, l= json.length; n < l; n++){
				data = json[n].data;
				if(data.$type == "crossBox"){
					data.$visible = deleteCreate;
					
				}
			}
			readOnly.deleteCreate = deleteCreate;
		}
		
		if(move != undefined){
			readOnly.move = move;
		}
	}
		
	/**
	 Adds a Node to the graph.
	 @param xPosition - the horizontal position of the added node. 0 is the center of the canvas' starting position
	 @param xPosition - the vertical position of the added node. 0 is the center of the canvas' starting position
	 @param nodeFamily - describes some properties of the node: (id, name, nodeClass, tooltip)
	 @param repositoryPorts - an array of properties for the input ports (id, name, tooltip)
	 @param inputPorts - an array of properties for the input ports (id, name, tooltip)
	 @param outputPorts - an array of properties for the output ports (id, name, tooltip)
	 @param nodeType - a string-identifier for a certain group of nodefamilies eg. Filter, Repository
	 @param forced - if true, no event listener will be called and the node will be created no matter what
	 */
	this.addNode = function(xPosition, yPosition, nodeFamily, repositoryPorts, inputPorts, outputPorts, nodeType, forced){
	
	// determine array size
	var countRepo = 0,
		countInput = 0,
		countOutput = 0,
		iconSpace = 0;
		
	if(repositoryPorts){
		countRepo = repositoryPorts.length;
		// check if it is an array
		if(!validArg(caller, repositoryPorts, 'object') || !countRepo){
			countRepo = 0;
		}
	}
	if(inputPorts){
		countInput = inputPorts.length;
		// check if it is an array
		if(!validArg(caller, inputPorts, 'object') || !countInput){
			countInput = 0;
		}
	}
	if(outputPorts){
		countOutput = outputPorts.length;
		// check if it is an array
		if(!validArg(caller, outputPorts, 'object') || !countOutput){
			countOutput = 0;
		}
	}
	
	// check for other illegal arguments
	var caller = 'addNode()';
	if(!validArg(caller, xPosition, 'number')){
		xPosition = 0;
	}
	if(!validArg(caller, yPosition, 'number')){
		yPosition = 0;
	}
	if(!validArg(caller, nodeType, 'string')){
		nodeType = 'Filter';
	}
	if(!validArgNode(caller, nodeFamily)){
		return;
	}
	
	// check if we have enough space reserved
	var max = jsonCapacity.max;
	var occ = jsonCapacity.occupied;
	if(occ == max){
		var families = json.slice(0, max);
		var ports = json.slice(max);
		
		for(var n = 0; n < max; n++){
			families.push(null);
		}
		max *= 2;
		jsonCapacity.max = max;
		json = families.concat(ports);
	}
	
	var maxPorts = Math.max(countInput, countRepo+countOutput+0.5),
		size = vertexLabelSize*2;
		
	// insert some space between repo ports and output ports, only if both are present
	if(maxPorts > countInput && (countRepo == 0 || countOutput == 0)){
		maxPorts-= 0.5;
	}
	
	// check if this nodeType has a defined icon
	if(nodeIcons[nodeType]){
		iconSpace = 6;
	}
		
	var width = (nodeFamily.name.length+4+iconSpace)*vertexLabelSize,
		height = size + 
			Math.max( 2*size, size/2 + maxPorts * size);
	
	// the classname may be longer than the diplayed name
	if( nodeFamily.nodeClass){
		width = Math.max(width,  (nodeFamily.nodeClass.length+6+iconSpace)*(vertexLabelSize-2));
	}
	
	// add big node box
	// change node color, depending on its type
	var nodeColor = nodeFillColor["nodeFamily"],
		strokeColor = nodeStrokeColor["nodeFamily"];
	var font = nodeFamily.font;
	if(!font){font = "Lucida Console"}
	
	var newNode = {
		  "data": {
			"$dim": size,					
			"$type": "nodeFamily",
			"$jsonIndex": occ,
			"$portCount": countRepo+countInput+countOutput,
			"$portIndex": json.length - max,
			"$icon": nodeIcons[nodeType],
			"$width": width,
			"$height": height,
			"$color": strokeColor,
			"$fillColor": nodeColor,
			"$nodeClass": nodeFamily.nodeClass,
			"$nodeType": nodeType,
			"$xPos": xPosition,
			"$yPos": yPosition,
			"$font": "px "+font,
			"$movable" : true,
			"$tooltip": nodeFamily.tooltip
		  }, 
		  "id": nodeFamily.id, 
		  "name": nodeFamily.name
	};
	json[occ] = newNode;
	
	// add closeButton
		var closeButton ={
			  "data": {
				"$dim": size,					
				"$type": "crossBox",
				"$family": newNode,
				"$xPos": xPosition+ width/2,
				"$yPos": yPosition- height/2,
				"$relativeX": width/2,
				"$relativeY": -height/2,
				"$color": strokeColor,
				"$fillColor": nodeColor,
				"$visible": readOnly.deleteCreate
			   }, 
			  "id": nodeFamily.id+".close", 
			  "name": "x"
		};
	json.push(closeButton);
	json[occ].data.$portCount++;
	
	// Loop Info: 
	// This loop adds all port types to the temporary port array.
	// For fast position computation, the relative position
	// of the first port to the nodeFamily-box is computed for each port type
	// it is stored in relativeX, relativeY
	
	for(var portType = 0; portType < 3; portType++){
		var x, y, relativeX, relativeY,
			loopPorts, loopType;
		switch(portType){
			case 0: // input ports
				relativeX = -width/2;
				relativeY = (0.75 - countInput/2 )*size;
				loopPorts = inputPorts;
				loopType = "inputPort";
				break;
		
			case 1: // repository ports
				var repoSpace = 0.5;
				if(countOutput == 0){
					repoSpace = 0;
				}
				relativeX = width/2;
				relativeY = (0.75 - (countOutput+countRepo+repoSpace)/2 )*size;
				loopPorts = repositoryPorts;
				loopType = "repositoryPort";
				break;
				
			case 2: // output ports
				var repoSpace = 0.5;
				if(countRepo == 0){
					repoSpace = 0;
				}
				relativeX = width/2;
				relativeY = (0.75 - (countOutput-countRepo-repoSpace)/2 )*size;
				loopPorts = outputPorts;
				loopType = "outputPort";
				break;
		}
		
		// continue only if ports exist
		if(loopPorts != null && loopPorts.length > 0){
		
			x = xPosition+ relativeX;
			y = yPosition+ relativeY;
			
			for(var p=0; p < loopPorts.length; p++){
				var port = loopPorts[p];
				
				// only attempt to add a node if it has an id
				if(validArgNode(caller, port)){
				
					var newPort ={
						  "adjacencies": [], 
						  "data": {
							"$dim": size,					
							"$type": loopType,
							"$xPos": x,
							"$yPos": y,
							"$color": nodeStrokeColor[loopType],
							"$fillColor": nodeColor,
							"$tooltip": port.tooltip,
							"$symbol": port.symbol}, 
						  "id": nodeFamily.id+"."+port.id, 
						  "name": port.name};
						  
					if(p == 0){
						newPort.data.$relativeX= relativeX;
						newPort.data.$relativeY= relativeY;
					}
					json.push(newPort);
					y += size;
				}
			}
		}
	}
	
		// call listener
		if(!forced){
			callListener("onCreateNode", [newNode]);
		}
		jsonCapacity.occupied++;
	}

	/**
		Removes a Node and all its attached edges from the graph.
		@param nodeFamily - the node which is to be removed
		@param forced - if true, the node will be removed unconditionally and no
						event listeners will be called
	*/
	this.removeNode =  function(nodeFamily, forced){
	
		// check for valid arguments
		if(!validArgNode("removeNode()", nodeFamily)){
			return;
		}
		// call listener
		if(!forced && !callListener("onRemoveNode", [nodeFamily])){
			return;
		}
		
		// unmark node
		if (markedNode && markedNode.id == nodeFamily.id){
			markedNode = null;
		}
		
		// removalFunction for custom nodes
		if(nodeFamily.data.$custom){
			var index = nodeFamily.data.$jsonIndex;
			delete json[index];
			json.splice(index, 1);
			return;
		}
		
		var deleteFrom = nodeFamily.data.$portIndex + jsonCapacity.max, 
			deleteUntil = deleteFrom+ nodeFamily.data.$portCount,
			familyIndex = nodeFamily.data.$jsonIndex,
			occ = --jsonCapacity.occupied; // decrement occupied spaces
		
		// delete nodeFamily and close button
		delete json[familyIndex];
		delete json[deleteFrom];
		
		// delete ports
		for(var n = deleteFrom+1;  n < deleteUntil;  n++){
			var portID = json[n].id;
			
			// remove all incoming edges
			for(var e = jsonCapacity.max+1, l = json.length;  e < l;  e++){
				var incNode = json[e];
				if(incNode){
					removeEdge(incNode.id, portID, true);	
				}
			}
			
			delete json[n];
		}
		
		// remove space that was used for ports
		var deletionSum = deleteUntil-deleteFrom; 
		json.splice(deleteFrom, deletionSum);
		selectedNode = null;
		
		// update nodeFamily indices and remove the deletion gap
		// in the array space by shifting all following families up
		var data;
		for(var n = familyIndex;  n < occ;  n++){
			json[n] = json[n+1];
			data = json[n].data;
			data.$jsonIndex--;
			data.$portIndex-= deletionSum;
			
		}
		json[occ] = null;
		
		// check if we should remove some space
		var max = jsonCapacity.max;
		if( (occ*4) < max && max > 10){
			var ports = json.slice(max);
			
			max /= 2;
			jsonCapacity.max = max;
			var families = json.slice(0, max);
			json = families.concat(ports);
		}
		return;
	 }
	 
	 /**
		Prepares a drag operation for a node, caching all graph nodes that need to be moved.
		This speeds up the moveNode() function considerably.
		@param node - the node which will be moved
	*/
	 function prepareNodeMove(node){
		
		var nodes = [];
		
		// memorize the nodeFamily
		nodes.push(node);
		
		// memorize all ports and the crossbox if it is a nodefamily
		if(!node.data.$custom){
			var	familyIndex = node.data.$jsonIndex,
				moveFrom = node.data.$portIndex+ jsonCapacity.max, 
				moveUntil = moveFrom+ node.data.$portCount,
				subnode;
			
			for(var i= moveFrom; i< moveUntil; i++){
				subnode = fd.graph.getNode(json[i].id);
				nodes.push(subnode);
			}
		}
		
		navi.draggedNodes = nodes;
	 }
	 
	 /**
		Moves a nodeFamily and its subnodes to the specified position.
		@param xPos - the horizontal position of the moved node. 0 is the center of the canvas' starting position
		@param yPos - the vertical position of the moved node. 0 is the center of the canvas' starting position
	 */
	 function moveNode(xPos, yPos){
		var nodes = navi.draggedNodes,
			node = nodes[0],
			dim = node.data.$dim;
		var x = xPos,
			y = yPos;
			
		// snap to grid
		if(grid.snap){
			var gridDim = grid.data.$dim,
				left = (xPos - node.data.$width/2),
				top = (yPos - node.data.$height/2),
				right = (xPos + node.data.$width/2),
				bottom = (yPos + node.data.$height/2),
				
				offLeft = Math.abs(left % gridDim),
				offTop = Math.abs(top % gridDim),
				offRight = Math.abs(right % gridDim),
				offBottom = Math.abs(bottom % gridDim),
				threshold = gridDim/6;
		
			if(top > 0){
				offTop = gridDim - offTop;
			}
			if(bottom > 0){
				offBottom = gridDim - offBottom;
			}
			if(left > 0){
				offLeft = gridDim - offLeft;
			}
			if(right > 0){
				offRight = gridDim - offRight;
			}
			
			// snap to closest gridline
			if(offLeft < threshold) {
				x += offLeft;
			}
			if(offRight < threshold) {
				x += offRight;
			}
			else if(offLeft > gridDim - threshold){
				x -= gridDim-offLeft;
			}
			else if(offRight > gridDim - threshold){
				x -= gridDim-offRight;
			}
			
			if(offTop < threshold) {
				y += offTop;
			}
			
			else if(offTop > gridDim - threshold){
				y -= gridDim-offTop;
			}
			
			else if(offBottom < threshold) {
				y += offBottom;
			}
			
			else if(offBottom > gridDim - threshold){
				y -= gridDim-offBottom;
			}
		}
	 
	 
		// set position of nodeFamily
		node.pos.setc(x, y);
		
		// set position of ports and crossbox
		var nodeX, nodeY;
		for(var i= 1, l = nodes.length; i < l; i++){
			node = nodes[i];
			
			if(node.data.$relativeX){
				nodeX = x+ node.data.$relativeX;
				nodeY = y+ node.data.$relativeY;
			}
			node.pos.setc(nodeX,nodeY);
			nodeY += dim;
		}
	 }
	 
	 /**
		Saves the position of a node, so that it will be placed there again,
		when the graph is refreshed. This function will only be called once
		when a node drag operation has finished.
	 */
	 function saveNodeMove(){
		 if(navi.draggedNodes){
			var node = navi.draggedNodes[0],
				familyIndex = node.data.$jsonIndex,
				pos = node.pos.getc(true);
			
			var data;
			if(familyIndex){
				data = json[familyIndex].data;
			}
			else{
				data = getNode(node.id).data;
			}
			
			// set pos of family box
			data.$xPos = pos.x;
			data.$yPos = pos.y;
			
			// move ports if it is a nodefamily
			if(!data.$custom){
				var moveFrom = node.data.$portIndex+ jsonCapacity.max-1, 
					nodes = navi.draggedNodes;
				
				// set pos of ports
				for(var i= 1, l = nodes.length; i < l; i++){
					data = json[moveFrom+i].data;
					pos = nodes[i].pos.getc(true);
					
					data.$xPos = pos.x;
					data.$yPos = pos.y;
				}
			}
			navi.draggedNodes = null;
		}
	 }

  /**
	Adds bendpoints to an edge.
	@param sourceID - the id of the node from which the edge starts
	@param targetID - the id of the node where the edge ends
	@param bendPoints - an array of {x,y} objects.
  */
  this.addBendPoints = function(sourceID, targetID, bendPoints){
		// check for valid arguments
	if(!validArg("addBendPoints()", sourceID, "string")){
		return;
	}
	if(!validArg("addBendPoints()", targetID, "string")){
		return;
	}
  
	// look up the source node and the nodeFamilys of both nodes
	var adja, loopNode;
	var sourceFamilyID = sourceID.substring(0, sourceID.indexOf("."));
		
	// look for nodefamilies
	for(var n = 0, l = jsonCapacity.occupied; n < l; n++){

		loopNode = json[n];
		if(loopNode.id == sourceFamilyID){
			adja = loopNode.adjacencies;
		}
	} 
	
	// nothing to delete?
	if(source == undefined){
		return;
	}
	var adja = source.adjacencies;
	
	for(var a=0; a< adja.length; a++){
		if(adja[a].nodeTo == targetID){
			delete adja[a];
			adja.splice(a,1);
			break;
		}
	}
	
  }
  
  /**
   *  Adds a directed edge between two nodes to the graph.
   *  @param sourceID - the id of the node from which the edge starts
   *  @param targetID - the id of the node where the edge ends
   *  @param edgeLabel - the label of the edge. Can be null if no label is required
   *  @param forced - if true, creates the edge unconditionally and does not trigger 
					  a listener event
	  @param noPlot - if true, the graph will not be plotted anew
   *  @return true - if the edge was successfully added
   */
	this.addEdge = function(sourceID, targetID, edgeLabel, forced, noPlot){
		// check for valid arguments
		if(!validArg("addEdge()", sourceID, "string")){
			return false;
		}
		if(!validArg("addEdge()", targetID, "string")){
			return false;
		}
		if(edgeLabel && !validArg("addEdge()", edgeLabel, "string")){
			edgeLabel = null;
		}
		
		// look up the source node and the nodeFamilys of both nodes
		var source, sourceFamily, targetFamily, target;
		
		var sourceFamilyID = sourceID.substring(0, sourceID.indexOf(".")), 
			targetFamilyID = targetID.substring(0, targetID.indexOf(".")),
			loopNode;
		// look for nodefamilies
		for(var n = 0, l = jsonCapacity.occupied; n < l; n++){
	
			loopNode = json[n];
			
			if(loopNode.id == sourceFamilyID){
				sourceFamily = loopNode
			}
			if(loopNode.id == targetFamilyID){
				targetFamily = loopNode
			}
		} 
		
		// look for ports
		for(var n = jsonCapacity.max, l = json.length; n < l; n++){
			loopNode = json[n];
			
			if(loopNode.id == sourceID){
				source = loopNode;
			}
			if(loopNode.id == targetID){
				target = loopNode;
			}
		}
		
		// the standard data set, which may be manipulated by event listeners
		var data = {};
		
		// call listener if the edge is not connected to the dummy node and check for permission
		if(!forced && sourceID != mouseNode.id && targetID != mouseNode.id){
		
			// check if ports and families were found
			if(!((sourceFamily || source.data.$custom) && source)){
				callListener("onError", ["Error in function addEdge(): The source port '" + sourceID + "' does not exist."]);
				return false;
			}
			if(!((targetFamily || target.data.$custom) && target)){
				callListener("onError", ["Error in function addEdge(): The target port '" + targetID + "' does not exist."]);
				return false;
			}
		
			if (!callListener("onCreateEdge", [sourceFamily, targetFamily, source, target, data])){
				return false;
			}
		}
		
		// if edge is connected to mouse, there is no targetFamily...
		var targetFamilyID = null;
		if(targetFamily){
			targetFamilyID = targetFamily.id;
		}
		
		var adja = source.adjacencies;
		// source node does not exist?
		if(!adja){
			return false;
		}
	
		var edge ={
				  "nodeTo": targetID, 
				  "nodeFrom": sourceID,
				  "data": {"$direction" : [ sourceID, targetID ],
						   "$targetFamily" : targetFamilyID,
						   "$color" : edgeColor}
				  };
		
		// change data if it was specified by the event listener
		for(var d in data){
			edge.data[d] = data[d];
		}
		
		// add label
		if(edgeLabel){
			edge.data.$label = edgeLabel;
        }
		
		// mouse edge
		if(sourceID == mouseNode.id || targetID == mouseNode.id){
			edge.data.$type = "mouseArrow";
		}
		
		adja.push(edge);
		
		// apply changes immediately
		
		if(!noPlot){ 
			var domSource = fd.graph.getNode(sourceID),
			domTarget = fd.graph.getNode(targetID);
			fd.graph.addAdjacence(domSource, domTarget, edge.data);
			fd.plot();
		}
		
		return true;
	}
  
  
  /**
		Sets or removes BendPoints of an Edge.
		@param sourceID - the ID of the connected outputPort
		@param targetID - the ID of the connected inputPort
		@param bendPoints - an array of {"x":..., "y":...} objects
		@param noPlot - if true, the graph will not be plotted anew
  */
  this.setBendPoints = function(sourceID, targetID, bendPoints, noPlot){
		// check for valid arguments
		if(!validArg("setBendPoints()", sourceID, "string")){
			return false;
		}
		if(!validArg("setBendPoints()", targetID, "string")){
			return false;
		}
		
		var adja = getNode(sourceID).adjacencies;
		// look for the specific edge
		for(var a = 0, l = adja.length; a < l; a++){
			if(adja[a].nodeTo == targetID){
				adja[a].data.$bendPoints = bendPoints;
				
				// update displayed graph (no need for refresh then)
				if(!noPlot){
					fd.graph.getAdjacence(sourceID, targetID).data.$bendPoints = bendPoints;
				}
				break;
			}
		}
		
		// plot visual changes
		if(!noPlot){ fd.plot(); }
  }
  
  /**
	Removes the edge between two nodes.
	@param sourceID - the id of the source Node
	@param targetID - the id of the target Node
	@param forced - if true, removes the edge unconditionally and
					does not trigger a listener event
	@param noPlot - if true, the graph will not be plotted anew
	*/
  this.removeEdge = function(sourceID, targetID, forced, noPlot){
	// check for valid arguments
	if(!validArg("removeEdge()", sourceID, "string")){
		return;
	}
	if(!validArg("removeEdge()", targetID, "string")){
		return;
	}
  
	// look up the source node and the nodeFamilys of both nodes
	var source, sourceFamily, targetFamily, target;
	
	var sourceFamilyID = sourceID.substring(0, sourceID.indexOf(".")), 
		targetFamilyID = targetID.substring(0, targetID.indexOf(".")),
		loopNode;
		
	// look for ports
	for(var n = jsonCapacity.max, l = json.length; n < l; n++){
		loopNode = json[n];
		if(loopNode){
			if(loopNode.id == sourceID){
				source = loopNode;
			}
			if(loopNode.id == targetID){
				target = loopNode;
			}
		}
	}
		
	if(!forced){
		// look for nodefamilies
		for(var n = 0, l = jsonCapacity.occupied; n < l; n++){

			loopNode = json[n];
			
			if(loopNode.id == sourceFamilyID){
				sourceFamily = loopNode
			}
			if(loopNode.id == targetFamilyID){
				targetFamily = loopNode
			}
		} 
	
	// call listener and ask for permission to delete
	if(sourceID != mouseNode.id && targetID != mouseNode.id){
			if(!callListener("onRemoveEdge", [sourceFamily, targetFamily, source, target])){
				return;
			}
		}
	}
	
	// nothing to delete?
	if(!source || !source.adjacencies){
		return;
	}
	var adja = source.adjacencies;
	
	for(var a=0; a< adja.length; a++){
		if(adja[a].nodeTo == targetID){
			delete adja[a];
			adja.splice(a,1);
			break;
		}
	}
	
	// apply changes immediately
	
	if(!noPlot){ 
		fd.graph.removeAdjacence(sourceID, targetID);
		fd.plot();
	}
  }
  
  /**
  Returns a node from the json array
  @param nodeID - the id of the desired node
  @return - the node object or null if the node does not exist
  */
  this.getNode = function(nodeID){
	// check for valid arguments
	if(!validArg("getNode()", nodeID, "string")){
		return;
	}
	for(var n=0, l = json.length; n < l; n++){
		var node = json[n];
		if(node && node.id == nodeID){
			return json[n];
		}
	}
	return null;
  }
  
  
  
  /**
	Highlights exactly one node or edge, and un-highlights the previously highlighted
	node or edge.
	@param node - the node or edge object which will be highlighted
	*/
  function setHighlight(node){
		// animation parameters
		var trans = $jit.Trans.Quint.easeOut, 
			dur= 200;
		// clean up the old highlight
		if(hover != null){
			
			if(hover.nodeFrom != undefined){
				hover.setData('lineWidth', 1, 'end');
				hover.setData('color', edgeColor, 'end');
			}
			else{
				var type = hover.data.$type;
				if(markedNode != null && hover.id == markedNode.id){
					hover.setData('color', markedNode.data.$color, 'end');
				}
				else{
					hover.setData('color', nodeStrokeColor[type], 'end');
				}
				// decrease size of ports and cross button
				if(type == 'inputPort' || type == 'outputPort' || type == 'repositoryPort'){
					hover.setData('dim', vertexLabelSize*2, 'end');
				}
				
			}
			hover = null;
		}
		
		// do we highlight something new?
		if(node != null){
			if(node.nodeFrom != undefined){
				node.setData('color', edgeColorFocus, 'end');
				node.setData('lineWidth', 2, 'end');
			}else{
				node.setData('color', nodeColorFocus, 'end');
				
				// increase size of ports and cross button
				var type = node.data.$type;
				if(type == 'inputPort' || type == 'outputPort' || type == 'repositoryPort'){
					node.setData('dim', vertexLabelSize*3, 'end');
				}
			}
			hover = node;
		}else{
			// slower fade away transition
			trans = $jit.Trans.Quint.easeIn;
			dur = 400;
		}
		
		// Animations overwrite each other. If we let this animation
		// execute while we delete an edge, the onComplete() function
		// of the edge animation will not be called
		fd.animate({  
				modes: ['edge-property:lineWidth:color',
						'node-property:color:dim'],
				transition: trans,  
				duration: dur
		}); 
  }
  
  /**
	Highlights exactly one node, and un-highlights the previously highlighted
	node.
	@param node - the node or edge object which will be highlighted.
				  Can be null if no new node shall be highlighted
	@param strokeColor - the color of the node stroke or a port's fillcolor.
						 Can be null for no change.
	@param fillColor - the fill color of a nodefamily
						Can be null for no change.
	*/
  this.markNode = function(node, strokeColor, fillColor){
  
		// check for valid arguments
		if(node && !validArgNode("markNode()", node)){
			return;
		}
		if(strokeColor && !validArgColor("markNode()", strokeColor)){
			return;
		}
		if(fillColor && !validArgColor("markNode()", fillColor)){
			return;
		}
		
		// clean up the old marking
		if(markedNode != null){
			
			if(markedNode.nodeFrom != undefined){
				// ignore edges
			}
			else{
				var type = markedNode.data.$type;
				markedNode.data.$color = nodeStrokeColor[type];
				
				// change FillColor only if it exists
				if(nodeFillColor[type]){
					markedNode.data.$fillColor = nodeFillColor[type];
				}
			}
			markedNode = null;
		}
		
		// do we highlight something new?
		if(node && node != null){
			// convert dom node to json node
		
			if(node.nodeFrom != undefined){
				// ignore edges
			}else{
				var jsonNode = getNode(node.id);
				if(!jsonNode){
					callListener("onError", ["Error in function markNode(): Marked Node was deleted."]);
					return;
				}
				
				if(strokeColor != null){
					jsonNode.data.$color = strokeColor;
				}
				
				// change FillColor only if it exists
				if(fillColor != null  && nodeFillColor[node.data.$type]){
					jsonNode.data.$fillColor = fillColor;
				}
			}
			markedNode = jsonNode;
		}
  }
  
  /**
   * @returns true if a mouse edge exists
   */
  this.isEdgeDragged = function(){
	  return !!selectedNode;
  }
  
  /**
   * Creates a dangling edge between a node and the mouse pointer.
   * If such an edge exists already, create an edge between both nodes
   * that were passed to this function.
   * 
   * @param node - the node which is to be connected to the mouse pointer
   * @param isIncoming - if true, the edge will point towards the node
   * @param label - an optional edge label
   */
  this.connectEdge = function(nodeID, isIncoming, label){
	
	// second node clicked?
	if(selectedNode){
		var label = selectedNode.label;
		
		// do we create an edge or simply discard it
		if(nodeID){
			// determine direction of edge
			if(selectedNode.isIncoming){
				addEdge(nodeID, selectedNode.id, label);
			}
			else{
				addEdge(selectedNode.id, nodeID, label);
			}
		}
		
		// remove mouse edge
		if(selectedNode.isIncoming){
			removeEdge(mouseNode.id, selectedNode.id, true);
		}else{
			removeEdge(selectedNode.id, mouseNode.id, true);
		}
		
		selectedNode = null;
	}
	// first node clicked
	else{
		selectedNode = {"id": nodeID, "isIncoming" : isIncoming, "label" : label};
		
		// determine direction of mouse edge
		if(isIncoming){
			addEdge(mouseNode.id, nodeID, label);
		}else{
			addEdge(nodeID, mouseNode.id, label);
		}
	}
  }
    
  /**
	Initializes the Graph, enabling canvas navigation and tooltips.
	Also the json array is initialized with a dummy mouse node and
	what happens on various mouse events is determined.
	*/
  this.initGraph = function(){
  
	if(!json){
		json = [ null, null, null, null, null, null, null, null, null, null, 
				{
				  "adjacencies": [], 
				  "data": {
					"$dim": 0,
					"$type": "none",
					"&xPos": 0,
					"&typeSelected" : null,
					"&yPos": 0
				  }, 
				  "id": "#DUMMY_MOUSE_NODE", 
				  "name": "",
				  "alpha": 0
			}];
	}
	  
	  fd = new $jit.FlowGraph({
		width : grid.data.$width,
		height : grid.data.$height,
	  
		//id of the visualization container
		injectInto: visContainer,
		//Enable zooming and panning
		//with scrolling and DnD
		Navigation: {
		  enable: true,
		  //Enable panning events only if we're dragging the empty
		  //canvas (and not a node).
		  panning: 'avoid nodes',
		  zooming: 40 //zoom speed. higher is more sensible
		},
		// Change node and edge styles such as
		// color and width.
		// These properties are also set per node
		// with dollar prefixed data-properties in the
		// JSON structure.
		Node: {
		  overridable: true,
		  CanvasStyles: {  
			  lineWidth: 2  
			}  
		  
		},
		Edge: {
		  overridable: true,
		  color: edgeColor,
		  lineWidth: 1,
		  dim: vertexLabelSize,
		  type: 'flowArrow'
		},
		//Native canvas text styling  
		Label: {  
			type: 'HTML', //Native or HTML  
			size: 10,  
			style: 'bold'  
		}, 
		//Add Tips  
		  Tips: {  
			enable: true,  
			type: 'Native',
			onShow: function(tip, node) {  
			  var message = node.data.$tooltip;
			  
			  // do not show a tooltip if it is undefined
			  if(message == undefined){
				tip.style.visibility = "hidden";
			  }
			  else{
				tip.innerHTML = "<div class=\"tip-text\">" + message + "</div>";  
				tip.style.visibility = "visible";
			  }
			}  
		 },  
		// Add node events
		Events: {
		  enable: true,
		  enableForEdges: true,
		  type: 'Native',
		  
		  /**
		   *	EVENT:	OnRightClick
		   */
		  onRightClick: function(node, eventInfo, e) {
			callListener("onRightClick", [node, eventInfo, e]);
		  },
		  
		  /**
		   *	EVENT:	OnDragCancel
		   */
		  onDragCancel: function(node, eventInfo, e) {
			callListener("onDragCancel", [node, eventInfo, e]);
		  },
		  
		  /**
		   *	EVENT:	OnMouseEnter
		   */
		  onMouseEnter: function(node, eventInfo, e) {
			if(node.id == mouseNode.id){
				return;
			}
			
			// ignore some highlights if we are dragging an edge
			if(selectedNode != null && 
				(node.data.$type == "nodeFamily" 
				|| node.data.$type == "crossBox")){
				return;
			}
			
			// check if object is an edge or a node
			mouseOverNode = (node.nodeFrom == undefined);
			mouseOverEdge = !mouseOverNode;
			
			setHighlight(node);
			if(readOnly.move && node.data.$type == "nodeFamily"){
				fd.canvas.getElement().style.cursor = 'move';
			}else if (readOnly.deleteCreate){
				fd.canvas.getElement().style.cursor = 'pointer';
			}
			callListener("onMouseEnter", [node, eventInfo, e]);
		  },
		  
		  /**
		   *	EVENT:	OnMouseMove
		   */
		  onMouseMove: function(node, eventInfo, e) {
		  /*if(!node && navi.isDragging){
			var mousePos = eventInfo.getPos(),
				deltaX = mousePos.x - navi.dragX,
				deltaY = mousePos.y - navi.dragY;
			fd.canvas.translate(deltaX, deltaY);
			deltaX = mousePos.x;
			deltaY = mousePos.y;
		  }
		  */
			// are we dragging an edge with the mouse ?
			if(selectedNode){
				var pos = eventInfo.getPos();
				mouseNode.pos.setc(pos.x, pos.y-2);
				fd.plot();
			}
			callListener("onMouseMove", [node, eventInfo, e]);
		  },
		  
		  /**
		   *	EVENT:	OnMouseLeave
		   */
		  onMouseLeave: function(node, eventInfo, e) {
			fd.canvas.getElement().style.cursor = '';
			mouseOverNode = false;
			mouseOverEdge = false;
			
			setHighlight(null);
			callListener("onMouseLeave", [node, eventInfo, e]);
		  },
		  
		  /**
		   *	EVENT:	OnDragStart
		   *	Actually this Event fires whenever the mouse button is down
		   */
		  onDragStart: function(node, eventInfo, e) {
		  var mousePos = eventInfo.getPos();
		  
		  // dragging the canvas
			if(node == false || node.id == mouseNode.id ||
			   !node.data.$movable){
				
			    navi.dragX = e.layerX//e.pageX;
				navi.dragY = e.layerX//e.pageY;
				navi.isDragging = true;
				return;
			}
			
			//dragging a node
			if(!selectedNode){
				// memorize where we drag the node instead on centering
				// it to the mouse pointer
				var	nodePos = node.pos.getc(true);
				
				// memorize the node position before node movement
				navi.dragX = nodePos.x;
				navi.dragY = nodePos.y;
				// the offset between the center of the node and the actual position
				// where we grabbed the node
				navi.nodeOffsetX = nodePos.x - mousePos.x;
				navi.nodeOffsetY = nodePos.y - mousePos.y;
				
				// prepare borders. These determine when to move the canvas with the node
				navi.borderLeft = (node.data.$width / 2 - navi.nodeOffsetX) * fd.canvas.scaleOffsetX;
				navi.borderTop = (node.data.$height / 2 - navi.nodeOffsetY) * fd.canvas.scaleOffsetX;
				navi.borderRight = grid.data.$width - (node.data.$width / 2 + navi.nodeOffsetX) * fd.canvas.scaleOffsetX;
				navi.borderBottom = grid.data.$height - (node.data.$height / 2 + navi.nodeOffsetY) * fd.canvas.scaleOffsetX;
				
				// we memorize all subnodes of the dragged nodefamily, to speed up
				// dragging
				prepareNodeMove(node);
			}
			callListener("onDragStart", [node, eventInfo, e]);
		  },
		  
		  /**
		   *	EVENT:	OnDragMove
		   */
		  onDragMove: function(node, eventInfo, e) {
			if(selectedNode || !readOnly.move || !node || 
				node.id == mouseNode.id || !node.data.$movable){
					return;
			}
			
			// move node
			var pos = eventInfo.getPos(),
				x = pos.x + navi.nodeOffsetX,
				y = pos.y + navi.nodeOffsetY;
			moveNode(x, y);
			
			// move screen if near edge
			var screenX = e.layerX,
				screenY = e.layerY;
				
			if(screenX < navi.borderLeft && x < navi.dragX){
				fd.canvas.translate(20,0);
			}
			else if(screenX > navi.borderRight && x > navi.dragX){
				fd.canvas.translate(-20,0);
			}
			
			if(screenY < navi.borderTop && y < navi.dragY){
				fd.canvas.translate(0,20);
			}
			else if(screenY > navi.borderBottom && y > navi.dragY){
				fd.canvas.translate(0,-20);
			}
			callListener("onDragMove", [node, eventInfo, e]);
			fd.plot();
			
		  },
		  
		  /**
		   *	EVENT:	OnDragEnd
		   */
		  onDragEnd: function(node, eventInfo, e) {
			if(node.data.$movable){
				saveNodeMove();
			}
			
			fd.plot();
			callListener("onDragEnd", [node, eventInfo, e]);
		  },
		  
		  /**
		   *	EVENT:	OnTouchMove
		   */
		  onTouchMove: function(node, eventInfo, e) {
			$jit.util.event.stop(e); //stop default touchmove event
			this.onDragMove(node, eventInfo, e);
		  },
		  
		  /**
		   *	EVENT:	OnClick
		   */
		  onClick: function(node, eventInfo, e){
		  
		  // finished dragging the screen
		  if(!node){
			var deltaX = e.layerX - navi.dragX,
				deltaY = e.layerY - navi.dragY;
			navi.centerX -= deltaX;
			navi.centerY -= deltaY;
			
			// remove selection and mouse-edge upon clicking anywhere
			  if(selectedNode){
				  connectEdge(null);
			  }
		  }
		  else if(mouseOverNode && readOnly.deleteCreate){
			
			var incomingEdge = false;
			switch(node.data.$type){
				case "crossBox":
					removeNode(node.data.$family);
					refresh();
					break;
					
				case "nodeFamily":
					break;
					
				case "inputPort":
					incomingEdge = true;
				case "outputPort":
				case "repositoryPort":
					connectEdge(node.id, incomingEdge);
					// set mouse node position
					if(selectedNode){
						var pos = node.pos.getc(true);
						// set mouse position
						mouseNode.pos.setc(pos.x, pos.y);
						fd.plot();
					}
					break;
			}
			
		  }
		  else if(mouseOverEdge && readOnly.deleteCreate){
			// no selection yet
				if(selectedNode == null){
					
					var source = node.data.$direction[0],
						label = node.data.$label;
					
					// remove selectedEdge
					removeEdge(source, node.data.$direction[1]);
					
					// create dangling edge
					connectEdge(source, false, label);
					
					// set mouse position
					var pos = eventInfo.getPos();
					mouseNode.pos.setc(pos.x, pos.y-2);
					fd.plot();
					
					return;
				}
		  }
		  
		  callListener("onClick", [node, eventInfo, e]);
		  navi.isDragging = false;
		  },
		  
		  /**
		   * 	EVENT: OnMouseWheel
		   */
		  onMouseWheel: function(e, delta) {
			  // update whether labels should be displayed
			  fd.canvas.showLabels = (fd.canvas.scaleOffsetX > fd.canvas.labelThreshold);
		  }
		},
		
		
		//Number of iterations for the FD algorithm
		iterations: 0,
		//Edge length
		levelDistance: 0
	  });
	  
	  // load JSON data.
	  refresh();
	  
	  // set zoom threshold for displaying text
	  fd.canvas.labelThreshold = 1/vertexLabelSize;
	  fd.canvas.showLabels = true;
	}
	
	// initialize graph
	initGraph();
	return this;
}